"""
Model Info and Parameter Tables
===============================

Defines :class:`ModelInfo` and :class:`ParameterTable` and the routines for
manipulating them.  In particular, :func:`make_model_info` converts a kernel
module into the model info block as seen by the rest of the sasmodels library.
"""
from __future__ import print_function

from copy import copy
from os.path import abspath, basename, splitext
import inspect
import logging

import numpy as np  # type: ignore

# Optional typing
# pylint: disable=unused-import
try:
    from typing import (
        Tuple, List, Union, Dict, Optional, Any, Callable, Sequence, Set,
        Mapping)
    from types import ModuleType
    Limits = Tuple[float, float]
    #LimitsOrChoice = Union[Limits, Tuple[Sequence[str]]]
    ParameterDef = Tuple[str, str, float, Limits, str, str]
    ParameterSetUser = Mapping[str, Union[float, List[float]]]
    ParameterSet = Mapping[str, float]
    TestInput = Union[str, float, List[float], Tuple[float, float], List[Tuple[float, float]]]
    TestValue = Union[float, List[float]]
    TestCondition = Tuple[ParameterSetUser, TestInput, TestValue]
except ImportError:
    pass
# pylint: enable=unused-import

logger = logging.getLogger(__name__)

# If MAX_PD changes, need to change the loop macros in kernel_iq.c
MAX_PD = 5 #: Maximum number of simultaneously polydisperse parameters

# assumptions about common parameters exist throughout the code, such as:
# (1) kernel functions Iq, Iqxy, Iqac, Iqabc, form_volume, ... don't see them
# (2) kernel drivers assume scale is par[0] and background is par[1]
# (3) mixture models drop the background on components and replace the scale
#     with a scale that varies from [-inf, inf]
# (4) product models drop the background and reassign scale
# and maybe other places.
# Note that scale and background cannot be coordinated parameters whose value
# depends on the some polydisperse parameter with the current implementation
DEFAULT_BACKGROUND = 1e-3
COMMON_PARAMETERS = [
    ("scale", "", 1, (0.0, np.inf), "", "Scale factor or Volume fraction"),
    ("background", "1/cm", DEFAULT_BACKGROUND, (-np.inf, np.inf), "", "Source background"),
]
NUM_COMMON_PARS = 2
NUM_MAGFIELD_PARS = 4
NUM_MAGNETIC_PARS = 3  # per sld
assert (len(COMMON_PARAMETERS) == NUM_COMMON_PARS
        and COMMON_PARAMETERS[0][0] == "scale"
        and COMMON_PARAMETERS[1][0] == "background"), "don't change common parameters"


def make_parameter_table(pars):
    # type: (List[ParameterDef]) -> ParameterTable
    """
    Construct a parameter table from a list of parameter definitions.

    This is used by the module processor to convert the parameter block into
    the parameter table seen in the :class:`ModelInfo` for the module.
    """
    processed = []
    for p in pars:
        if not isinstance(p, (list, tuple)) or len(p) != 6:
            raise ValueError("Parameter should be [name, units, default, limits, type, desc], but got %r"
                             %str(p))
        processed.append(parse_parameter(*p))
    partable = ParameterTable(processed)
    partable.check_angles(strict=True)
    return partable

def parse_parameter(name, units='', default=np.nan,
                    user_limits=None, ptype='', description=''):
    # type: (str, str, float, Sequence[Any], str, str) -> Parameter
    """
    Parse an individual parameter from the parameter definition block.

    This does type and value checking on the definition, leading
    to early failure in the model loading process and easier debugging.
    """
    # Parameter is a user facing class.  Do robust type checking.
    if not isstr(name):
        raise ValueError("expected string for parameter name %r"%name)
    if not isstr(units):
        raise ValueError("expected units to be a string for %s"%name)

    # Process limits as [float, float] or [[str, str, ...]]
    choices = []  # type: List[str]
    if user_limits is None:
        limits = (-np.inf, np.inf)
    elif not isinstance(user_limits, (tuple, list)):
        raise ValueError("invalid limits for %s"%name)
    else:
        # if limits is [[str,...]], then this is a choice list field,
        # and limits are 1 to length of string list
        if isinstance(user_limits[0], (tuple, list)):
            choices = user_limits[0]
            limits = (0., len(choices)-1.)
            if not all(isstr(k) for k in choices):
                raise ValueError("choices must be strings for %s"%name)
        else:
            try:
                low, high = user_limits
                limits = (float(low), float(high))
            except Exception:
                raise ValueError("invalid limits for %s: %r"%(name, user_limits))
            if low >= high:
                raise ValueError("require lower limit < upper limit")

    # Process default value as float, making sure it is in range
    if not isinstance(default, (int, float)):
        raise ValueError("expected default %r to be a number for %s"
                         % (default, name))
    if default < limits[0] or default > limits[1]:
        raise ValueError("default value %r not in range for %s"
                         % (default, name))

    # Check for valid parameter type
    if ptype not in ("volume", "orientation", "sld", "magnetic", ""):
        raise ValueError("unexpected type %r for %s" % (ptype, name))

    # Check for valid parameter description
    if not isstr(description):
        raise ValueError("expected description to be a string")

    # Parameter id for name[n] does not include [n]
    ref = None  # type: Optional[str]
    if "[" in name:
        if not name.endswith(']'):
            raise ValueError("Expected name[len] for vector parameter %s"%name)
        pid, ref = name[:-1].split('[', 1)
        ref = ref.strip()
    else:
        pid, ref = name, None

    # automatically identify sld types
    if ptype == '' and (pid.startswith('sld') or pid.endswith('sld')):
        ptype = 'sld'

    # Check if using a vector definition, name[k], as the parameter name
    if ref:
        if ref == '':
            raise ValueError("Need to specify vector length for %s"%name)
        try:
            length = int(ref)
            control = None
        except ValueError:
            length = None
            control = ref
    else:
        length = 1
        control = None

    # Build the parameter
    parameter = Parameter(name=name, units=units, default=default,
                          limits=limits, ptype=ptype, description=description)

    # TODO: need better control over whether a parameter is polydisperse
    parameter.polydisperse = ptype in ('orientation', 'volume')
    parameter.relative_pd = ptype == 'volume'
    parameter.choices = choices
    parameter.length = length
    parameter.length_control = control
    return parameter


def expand_pars(partable, pars=None):
    # type: (ParameterTable, ParameterSetUser) ->  ParameterSet
    """
    Create a parameter set from key-value pairs.

    *pars* are the key-value pairs to use for the parameters.  Any
    parameters not specified in *pars* are set from the *partable* defaults.

    If *pars* references vector fields, such as thickness[n], then support
    different ways of assigning the parameter values, including assigning a
    specific value (e.g., thickness3=50.0), assigning a new value to all
    (e.g., thickness=50.0) or assigning values using list notation.
    """
    if pars is None:
        result = partable.defaults
    else:
        lookup = dict((p.id, p) for p in partable.kernel_parameters)
        result = dict(partable.defaults)
        scalars = dict((name, value) for name, value in pars.items()
                       if name not in lookup or lookup[name].length == 1)
        vectors = dict((name, value) for name, value in pars.items()
                       if name in lookup and lookup[name].length > 1)
        #print("lookup", lookup)
        #print("scalars", scalars)
        #print("vectors", vectors)
        for name, value in vectors.items():
            if np.isscalar(value):
                # support for the form
                #    dict(thickness=0, thickness2=50)
                for k in range(1, lookup[name].length+1):
                    key = name+str(k)
                    if key not in scalars:
                        scalars[key] = value
            else:
                # supoprt for the form
                #    dict(thickness=[20,10,3])
                for (k, v) in enumerate(value):
                    scalars[name+str(k+1)] = v
        result.update(scalars)
        #print("expanded", result)

    return result

def prefix_parameter(par, prefix):
    # type: (Parameter, str) -> Parameter
    """
    Return a copy of the parameter with its name prefixed.
    """
    new_par = copy(par)
    new_par.name = prefix + par.name
    new_par.id = prefix + par.id
    return new_par

def suffix_parameter(par, suffix):
    # type: (Parameter, str) -> Parameter
    """
    Return a copy of the parameter with its name prefixed.
    """
    new_par = copy(par)
    # If name has the form x[n], replace with x_suffix[n]
    new_par.name = par.id + suffix + par.name[len(par.id):]
    new_par.id = par.id + suffix
    return new_par

class Parameter(object):
    """
    The available kernel parameters are defined as a list, with each parameter
    defined as a sublist with the following elements:

    *name* is the name that will be displayed to the user.  Names
    should be lower case, with words separated by underscore.  If
    acronyms are used, the whole acronym should be upper case. For vector
    parameters, the name will be followed by *[len]* where *len* is an
    integer length of the vector, or the name of the parameter which
    controls the length.  The attribute *id* will be created from name
    without the length.

    *units* should be one of *degrees* for angles, *Ang* for lengths,
    *1e-6/Ang^2* for SLDs.

    *default* will be the initial value for the model when it
    is selected, or when an initial value is not otherwise specified.

    *limits ([lb, ub])* are the hard limits on the parameter value, used to
    limit the polydispersity density function.  In the fit, the parameter limits
    given to the fit are the limits  on the central value of the parameter.
    If there is polydispersity, it will evaluate parameter values outside
    the fit limits, but not outside the hard limits specified in the model.
    If there are no limits, use +/-inf imported from numpy.

    *type* indicates how the parameter will be used.  "volume" parameters
    will be used in all functions.  "orientation" parameters are not passed,
    but will be used to convert from *qx*, *qy* to *qa*, *qb*, *qc* in calls to
    *Iqxy* and *Imagnetic*.  If *type* is the empty string, the parameter will
    be used in all of *Iq*, *Iqxy* and *Imagnetic*.  "sld" parameters
    can automatically be promoted to magnetic parameters, each of which
    will have a magnitude and a direction, which may be different from
    other sld parameters. The volume parameters are used for calls
    to form_volume within the kernel (required for volume normalization),
    to shell_volume (for hollow shapes), and to radius_effective (for
    structure factor interactions) respectively.

    *description* is a short description of the parameter.  This will
    be displayed in the parameter table and used as a tool tip for the
    parameter value in the user interface.

    Additional values can be set after the parameter is created:

    * *length* is the length of the field if it is a vector field

    * *length_control* is the parameter which sets the vector length

    * *is_control* is True if the parameter is a control parameter for a vector

    * *polydisperse* is true if the parameter accepts a polydispersity

    * *relative_pd* is true if that polydispersity is a portion of the
      value (so a 10% length dipsersity would use a polydispersity value
      of 0.1) rather than absolute dispersisity (such as an angle plus or
      minus 15 degrees).

    *choices* is the option names for a drop down list of options, as for
    example, might be used to set the value of a shape parameter.

    Control parameters are used for variant models such as :ref:`rpa` which
    have different cases with different parameters, as well as models
    like *spherical_sld* with its user defined number of shells.
    The control parameter should appear in the parameter table along with the
    parameters it is is controlling.  For variant models, use *[CASES]* in
    place of the parameter limits Within the parameter definition table,
    with case names such as::

         CASES = ["diblock copolymer", "triblock copolymer", ...]

    This should give *limits=[[case1, case2, ...]]*, but the model loader
    translates it to *limits=[0, len(CASES)-1]*, and adds *choices=CASES* to
    the :class:`Parameter` definition. Note that models can use a list of
    cases as a parameter without it being a control parameter.  Either way,
    the parameter is sent to the model evaluator as *float(choice_num)*,
    where choices are numbered from 0. :meth:`ModelInfo.get_hidden_parameters`
    will determine which parameers to display.

    The class contructor should not be called directly, but instead the
    parameter table is built using :func:`make_parameter_table` and
    :func:`parse_parameter` therein.
    """
    def __init__(self, name, units='', default=np.nan, limits=(-np.inf, np.inf),
                 ptype='', description=''):
        # type: (str, str, float, Limits, str, str) -> None
        self.id = name.split('[')[0].strip() # type: str
        self.name = name                     # type: str
        self.units = units                   # type: str
        self.default = default               # type: float
        self.limits = limits                 # type: Limits
        self.type = ptype                    # type: str
        self.description = description       # type: str

        # Length and length_control will be filled in once the complete
        # parameter table is available.
        self.length = 1                      # type: int
        self.length_control = None           # type: Optional[str]
        self.is_control = False              # type: bool

        # TODO: need better control over whether a parameter is polydisperse
        self.polydisperse = False            # type: bool
        self.relative_pd = False             # type: bool

        # choices are also set externally.
        self.choices = []                    # type: List[str]

    def as_definition(self):
        # type: () -> str
        """
        Declare space for the variable in a parameter structure.

        For example, the parameter thickness with length 3 will
        return "double thickness[3];", with no spaces before and
        no new line character afterward.
        """
        if self.length == 1:
            return "double %s;"%self.id
        else:
            return "double %s[%d];"%(self.id, self.length)

    def as_function_argument(self):
        # type: () -> str
        r"""
        Declare the variable as a function argument.

        For example, the parameter thickness with length 3 will
        return "double \*thickness", with no spaces before and
        no comma afterward.
        """
        if self.length == 1:
            return "double %s"%self.id
        else:
            return "double *%s"%self.id

    def __str__(self):
        # type: () -> str
        return "<%s>"%self.name

    def __repr__(self):
        # type: () -> str
        return "P<%s>"%self.name

class ParameterTable(object):
    """
    ParameterTable manages the list of available parameters.

    There are a couple of complications which mean that the list of parameters
    for the kernel differs from the list of parameters that the user sees.

    (1) Common parameters.  Scale and background are implicit to every model,
    but are not passed to the kernel.

    (2) Vector parameters.  Vector parameters are passed to the kernel as a
    pointer to an array, e.g., thick[], but they are seen by the user as n
    separate parameters thick1, thick2, ...

    Therefore, the parameter table is organized by how it is expected to be
    used. The following information is needed to set up the kernel functions:

    * *kernel_parameters* is the list of parameters in the kernel parameter
      table, with vector parameter p declared as p[].

    * *iq_parameters* is the list of parameters to the Iq(q, ...) function,
      with vector parameter p sent as p[].

    * *form_volume_parameters* is the list of parameters to the form_volume(...)
      function, with vector parameter p sent as p[].

    Problem details, which sets up the polydispersity loops, requires the
    following:

    * *theta_offset* is the offset of the theta parameter in the kernel parameter
      table, with vector parameters counted as n individual parameters
      p1, p2, ..., or offset is -1 if there is no theta parameter.

    * *max_pd* is the maximum number of polydisperse parameters, with vector
      parameters counted as n individual parameters p1, p2, ...  Note that
      this number is limited to sasmodels.modelinfo.MAX_PD.

    * *npars* is the total number of parameters to the kernel, with vector
      parameters counted as n individual parameters p1, p2, ...

    * *common_parameters* is the list of common parameters, with a unique
      copy for each model so that structure factors can have a default
      background of 0.0.

    * *call_parameters* is the complete list of parameters to the kernel,
      including scale and background, with vector parameters recorded as
      individual parameters p1, p2, ...

    * *active_1d* is the set of names that may be polydisperse for 1d data

    * *active_2d* is the set of names that may be polydisperse for 2d data

    User parameters are the set of parameters visible to the user, including
    the scale and background parameters that the kernel does not see.  User
    parameters don't use vector notation, and instead use p1, p2, ...
    """
    def __init__(self, parameters):
        # type: (List[Parameter]) -> None

        # scale and background are implicit parameters
        # Need them to be unique to each model in case they have different
        # properties, such as default=0.0 for structure factor backgrounds.
        self.common_parameters = [Parameter(*p) for p in COMMON_PARAMETERS]
        self.kernel_parameters = parameters
        self._set_vector_lengths()
        self.npars = sum(p.length for p in self.kernel_parameters)
        self.nmagnetic = sum(p.length for p in self.kernel_parameters
                             if p.type == 'sld')
        self.nvalues = NUM_COMMON_PARS + self.npars
        if self.nmagnetic:
            self.nvalues += NUM_MAGFIELD_PARS + NUM_MAGNETIC_PARS*self.nmagnetic
        self.call_parameters = self._get_call_parameters()
        self.defaults = self._get_defaults()
        #self._name_table= dict((p.id, p) for p in parameters)

        # Set the kernel parameters.  Assumes background and scale are the
        # first two parameters in the parameter list, but these are not sent
        # to the underlying kernel functions.
        self.iq_parameters = [p for p in self.kernel_parameters
                              if p.type not in ('orientation', 'magnetic')]
        self.orientation_parameters = [p for p in self.kernel_parameters
                                       if p.type == 'orientation']
        self.form_volume_parameters = [p for p in self.kernel_parameters
                                       if p.type == 'volume']

        # Theta offset
        offset = 0
        for p in self.kernel_parameters:
            if p.name == 'theta':
                self.theta_offset = offset
                break
            offset += p.length
        else:
            self.theta_offset = -1

        # number of polydisperse parameters
        num_pd = sum(p.length for p in self.kernel_parameters if p.polydisperse)
        # Don't use more polydisperse parameters than are available in the model
        self.max_pd = min(num_pd, MAX_PD)

        # true if has 2D parameters
        self.has_2d = any(p.type in ('orientation', 'magnetic')
                          for p in self.kernel_parameters)
        self.is_asymmetric = any(p.name == 'psi' for p in self.kernel_parameters)
        self.magnetism_index = [k for k, p in enumerate(self.call_parameters)
                                if p.id.endswith('_M0')]

        self.pd_1d = set(p.name for p in self.call_parameters
                         if p.polydisperse and p.type not in ('orientation', 'magnetic'))
        self.pd_2d = set(p.name for p in self.call_parameters if p.polydisperse)

        # Final checks
        self.check_duplicates()
        self.check_angles()

    def set_zero_background(self):
        # type: () -> None
        """
        Set the default background to zero for this model.  This is done for
        structure factor models.
        """
        # Make sure background is the second common parameter.
        assert self.common_parameters[1].id == "background"
        self.common_parameters[1].default = 0.0
        self.defaults = self._get_defaults()

    def check_angles(self, strict=False):
        """
        Check that orientation angles are theta, phi and possibly psi.

        *strict* should be True when checking a parameter table defined
        in a model file, but False when checking from mixture models, etc.,
        where the parameters aren't being passed to a calculator directly.
        """
        theta = phi = psi = -1
        for k, p in enumerate(self.kernel_parameters):
            if p.name == 'theta':
                theta = k
                if p.type != 'orientation':
                    raise TypeError("theta must be an orientation parameter")
            elif p.name == 'phi':
                phi = k
                if p.type != 'orientation':
                    raise TypeError("phi must be an orientation parameter")
            elif p.name == 'psi':
                psi = k
                if p.type != 'orientation':
                    raise TypeError("psi must be an orientation parameter")
            elif strict and p.type == 'orientation':
                raise TypeError("only theta, phi and psi can be orientation parameters")
        if theta >= 0 and phi >= 0:
            last_par = len(self.kernel_parameters) - 1
            if phi != theta+1:
                raise TypeError("phi must follow theta")
            if psi >= 0 and psi != phi+1:
                raise TypeError("psi must follow phi")
            # TODO: Why must theta/phi/psi be at the end?  Consistency only?
            if strict and phi != last_par and psi != last_par:
                raise TypeError("orientation parameters must appear at the "
                                "end of the parameter table")
        elif theta >= 0 or phi >= 0 or psi >= 0:
            raise TypeError("oriented shapes must have both theta and phi and maybe psi")

    def check_duplicates(self):
        """
        Check for duplicate parameter names
        """
        checked, dups = set(), set()
        for p in self.call_parameters:
            if p.id in checked:
                dups.add(p.id)
            else:
                checked.add(p.id)
        if dups:
            raise TypeError("Duplicate parameters: {}"
                            .format(", ".join(sorted(dups))))

    def __getitem__(self, key):
        # Find the parameter definition
        for par in self.call_parameters:
            if par.name == key:
                return par
        raise KeyError("unknown parameter %r"%key)

    def __contains__(self, key):
        for par in self.call_parameters:
            if par.name == key:
                return True
        return False

    def _set_vector_lengths(self):
        # type: () -> None
        """
        Walk the list of kernel parameters, setting the length field of the
        vector parameters from the upper limit of the reference parameter.

        This needs to be done once the entire parameter table is available
        since the reference may still be undefined when the parameter is
        initially created.

        Returns the list of control parameter names.

        Note: This modifies the underlying parameter object.
        """
        # Sort out the length of the vector parameters such as thickness[n]
        for p in self.kernel_parameters:
            if p.length_control:
                ref = self._get_ref(p)
                ref.is_control = True
                ref.polydisperse = False
                low, high = ref.limits
                if int(low) != low or int(high) != high or low < 0 or high > 20:
                    raise ValueError("expected limits on %s to be within [0, 20]"
                                     % ref.name)
                p.length = int(high)

    def _get_ref(self, p):
        # type: (Parameter) -> Parameter
        for ref in self.kernel_parameters:
            if ref.id == p.length_control:
                return ref
        raise ValueError("no reference variable %r for %s"
                         % (p.length_control, p.name))

    def _get_defaults(self):
        # type: () -> ParameterSet
        """
        Get a list of parameter defaults from the parameters.

        Expands vector parameters into parameter id+number.
        """
        # Construct default values, including vector defaults
        defaults = {}
        for p in self.call_parameters:
            if p.length == 1:
                defaults[p.id] = p.default
            else:
                for k in range(1, p.length+1):
                    defaults["%s%d"%(p.id, k)] = p.default
        return defaults

    def _get_call_parameters(self):
        # type: () -> List[Parameter]
        full_list = self.common_parameters[:]
        for p in self.kernel_parameters:
            if p.length == 1:
                full_list.append(p)
            else:
                for k in range(1, p.length+1):
                    pk = Parameter(p.id+str(k), p.units, p.default,
                                   p.limits, p.type, p.description)
                    pk.polydisperse = p.polydisperse
                    pk.relative_pd = p.relative_pd
                    pk.choices = p.choices
                    full_list.append(pk)

        # Add the magnetic parameters to the end of the call parameter list.
        if self.nmagnetic > 0:
            full_list.extend([
                Parameter('up_frac_i', '', 0., [0., 1.],
                          'magnetic', 'fraction of spin up incident'),
                Parameter('up_frac_f', '', 0., [0., 1.],
                          'magnetic', 'fraction of spin up final'),
                Parameter('up_theta', 'degrees', 90., [0., 360.],
                          'magnetic', 'polarization axis rotation angle'),
                Parameter('up_phi', 'degrees', 0., [0., 180.],
                          'magnetic', 'polarization axis inclination angle'),
            ])
            slds = [p for p in full_list if p.type == 'sld']
            for p in slds:
                full_list.extend([
                    Parameter(p.id+'_M0', '1e-6/Ang^2', 0., [-np.inf, np.inf],
                              'magnetic', 'magnetic amplitude for '+p.description),
                    Parameter(p.id+'_mtheta', 'degrees', 0., [-90., 90.],
                              'magnetic', 'magnetic latitude for '+p.description),
                    Parameter(p.id+'_mphi', 'degrees', 0., [-180., 180.],
                              'magnetic', 'magnetic longitude for '+p.description),
                ])
        #print("call parameters", full_list)
        return full_list

    def user_parameters(self, pars, is2d=True):
        # type: (Dict[str, float], bool) -> List[Parameter]
        """
        Return the list of parameters for the given data type.

        Vector parameters are expanded in place.  If multiple parameters
        share the same vector length, then the parameters will be interleaved
        in the result.  The control parameters come first.  For example,
        if the parameter table is ordered as::

            sld_core
            sld_shell[num_shells]
            sld_solvent
            thickness[num_shells]
            num_shells

        and *pars[num_shells]=2* then the returned list will be::

            num_shells
            scale
            background
            sld_core
            sld_shell1
            thickness1
            sld_shell2
            thickness2
            sld_solvent

        Note that shell/thickness pairs are grouped together in the result
        even though they were not grouped in the incoming table.  The control
        parameter is always returned first since the GUI will want to set it
        early, and rerender the table when it is changed.

        Parameters marked as sld will automatically have a set of associated
        magnetic parameters (p_M0, p_mtheta, p_mphi), as well as polarization
        information (up_theta, up_phi, up_frac_i, up_frac_f).
        """
        # control parameters go first
        control = [p for p in self.kernel_parameters if p.is_control]

        # Gather entries such as name[n] into groups of the same n
        dependent = {} # type: Dict[str, List[Parameter]]
        dependent.update((p.id, []) for p in control)
        for p in self.kernel_parameters:
            if p.length_control is not None:
                dependent[p.length_control].append(p)

        # Gather entries such as name[4] into groups of the same length
        fixed_length = {}  # type: Dict[int, List[Parameter]]
        for p in self.kernel_parameters:
            if p.length > 1 and p.length_control is None:
                fixed_length.setdefault(p.length, []).append(p)

        # Using the call_parameters table, we already have expanded forms
        # for each of the vector parameters; put them in a lookup table
        # Note: p.id and p.name are currently identical for the call parameters
        expanded_pars = dict((p.id, p) for p in self.call_parameters)

        def append_group(name):
            """add the named parameter, and related magnetic parameters if any"""
            result.append(expanded_pars[name])
            if is2d:
                for tag in '_M0', '_mtheta', '_mphi':
                    if name+tag in expanded_pars:
                        result.append(expanded_pars[name+tag])

        # Gather the user parameters in order
        result = control + self.common_parameters
        for p in self.kernel_parameters:
            if not is2d and p.type in ('orientation', 'magnetic'):
                pass
            elif p.is_control:
                pass # already added
            elif p.length_control is not None:
                table = dependent.get(p.length_control, [])
                if table:
                    # look up length from incoming parameters
                    table_length = int(pars.get(p.length_control, p.length))
                    del dependent[p.length_control] # first entry seen
                    for k in range(1, table_length+1):
                        for entry in table:
                            append_group(entry.id+str(k))
                else:
                    pass # already processed all entries
            elif p.length > 1:
                table = fixed_length.get(p.length, [])
                if table:
                    table_length = p.length
                    del fixed_length[p.length]
                    for k in range(1, table_length+1):
                        for entry in table:
                            append_group(entry.id+str(k))
                else:
                    pass # already processed all entries
            else:
                append_group(p.id)

        if is2d and 'up_theta' and 'up_phi' in expanded_pars:
            result.extend([
                expanded_pars['up_frac_i'],
                expanded_pars['up_frac_f'],
                expanded_pars['up_theta'],
                expanded_pars['up_phi'],
            ])

        return result

def _simple_insert(parameters, insert, remove):
    """
    Build new parameters from old with insertion locations specified.

    *parameters* is the existing parameter list.

    *insert* is list of parameters to insert.

    *remove* is list of parameter names to remove.

    The new parameters are inserted as a block replacing the first removed
    parameter.
    """
    remove = set(remove)
    new_list = []
    for par in parameters:
        if par.id not in remove:
            new_list.append(par)
        if par.id in remove and insert:
            new_list.extend(insert)
            insert = []
    return new_list

def _insert_after(parameters, insert, remove, insert_after):
    """
    Build new parameters from old with insertion locations specified.

    *parameters* is the existing parameter list.

    *insert* is list of parameters to insert.

    *remove* is list of parameter names to remove.

    *insert_after* where to insert names, as {"old": "new,new,..."}
    """
    remove = set(remove)
    new_list = []
    lookup = {p.id: p for p in insert}
    def _process_group(key):
        items = insert_after.get(key, None)
        if items is None:
            return
        for name in items.split(','):
            # Make sure par is available
            if name not in lookup:
                raise ValueError("variable %s not in new parameters"
                                 " when inserting %s after %r"
                                 % (name, items, key))
            # Make sure par hasn't be used
            if lookup[name] is None:
                raise ValueError("variable %s already processed"
                                 " when inserting %s after %r"
                                 % (name, items, key))
            # Append par
            new_list.append(lookup[name])
            # Mark par as used
            lookup[name] = None
    _process_group("")
    for par in parameters:
        if par.id not in remove:
            new_list.append(par)
        _process_group(par.id)
    for name, par in lookup.items():
        if par is not None:
            raise ValueError("parameter %s not listed in insert_after" % name)

    return new_list

def derive_table(table, insert, remove, insert_after=None):
    # type: (ParameterTable, List[str], List[Parameter], Optional[Dict[str, str]]) -> ParameterTable
    """
    Create a derived parameter table.

    Parameters given in *insert* are added to the table and parameters
    named in *remove* are deleted from the table. If *insert_after* is
    provided, then it indicates where in the new parameter table the
    parameters are inserted.
    """
    old = table.kernel_parameters
    if insert_after is None:
        new = _simple_insert(old, insert, remove)
    else:
        new = _insert_after(old, insert, remove, insert_after)
    return ParameterTable(new)

def isstr(x):
    # type: (Any) -> bool
    """
    Return True if the object is a string.
    """
    # TODO: 2-3 compatible tests for str, including unicode strings
    return isinstance(x, str)


#: Set of variables defined in the model that might contain C code
C_SYMBOLS = ['Imagnetic', 'Iq', 'Iqxy', 'Iqac', 'Iqabc',
             'form_volume', 'shell_volume', 'c_code', 'valid']

def _find_source_lines(model_info, kernel_module):
    # type: (ModelInfo, ModuleType) -> None
    """
    Identify the location of the C source inside the model definition file.

    This code runs through the source of the kernel module looking for lines
    that contain C code (because it is a c function definition).  Clearly
    there are all sorts of reasons why this might not work (e.g., code
    commented out in a triple-quoted line block, code built using string
    concatenation, code defined in the branch of an 'if' block, code imported
    from another file), but it should work properly in the 95% case, and for
    the remainder, getting the incorrect line number will merely be
    inconvenient.
    """
    # Only need line numbers if we are creating a C module and the C symbols
    # are defined.
    if (callable(model_info.Iq)
            or not any(hasattr(model_info, s) for s in C_SYMBOLS)):
        return

    # load the module source if we can
    try:
        source = inspect.getsource(kernel_module)
    except IOError:
        return

    # look for symbol at the start of the line
    for lineno, line in enumerate(source.split('\n')):
        for name in C_SYMBOLS:
            if line.startswith(name):
                # Add 1 since some compilers complain about "#line 0"
                model_info.lineno[name] = lineno + 1
                break


def make_model_info(kernel_module):
    # type: (ModuleType) -> ModelInfo
    """
    Extract the model definition from the loaded kernel module.

    Fill in default values for parts of the module that are not provided.

    Note: vectorized Iq and Iqac/Iqabc functions will be created for python
    models when the model is first called, not when the model is loaded.
    """
    if hasattr(kernel_module, "model_info"):
        # Custom sum/multi models
        return kernel_module.model_info

    info = ModelInfo()

    # Build the parameter table
    #print("make parameter table", kernel_module.parameters)
    parameters = make_parameter_table(getattr(kernel_module, 'parameters', []))

    # background defaults to zero for structure factor models
    structure_factor = getattr(kernel_module, 'structure_factor', False)
    if structure_factor:
        parameters.set_zero_background()

    filename = abspath(kernel_module.__file__).replace('.pyc', '.py')
    kernel_id = splitext(basename(filename))[0]
    name = getattr(kernel_module, 'name', None)
    if name is None:
        name = " ".join(w.capitalize() for w in kernel_id.split('_'))

    info.id = kernel_id  # string used to load the kernel
    info.basefile = info.filename = filename
    info.name = name
    info.title = getattr(kernel_module, 'title', name+" model")
    info.description = getattr(kernel_module, 'description', 'no description')
    info.base = info.parameters = parameters
    info.translation = None
    info.composition = None
    info.docs = kernel_module.__doc__
    info.category = getattr(kernel_module, 'category', None)
    info.structure_factor = getattr(kernel_module, 'structure_factor', False)
    # TODO: find Fq by inspection
    info.radius_effective_modes = getattr(kernel_module, 'radius_effective_modes', None)
    info.have_Fq = getattr(kernel_module, 'have_Fq', False)
    info.profile_axes = getattr(kernel_module, 'profile_axes', ['x', 'y'])
    # Note: custom.load_custom_kernel_module assumes the C sources are defined
    # by this attribute.
    info.source = getattr(kernel_module, 'source', [])
    info.c_code = getattr(kernel_module, 'c_code', None)
    info.radius_effective = getattr(kernel_module, 'radius_effective', None)
    # TODO: check the structure of the tests
    info.tests = getattr(kernel_module, 'tests', [])
    info.valid = getattr(kernel_module, 'valid', '')
    info.form_volume = getattr(kernel_module, 'form_volume', None) # type: ignore
    info.shell_volume = getattr(kernel_module, 'shell_volume', None) # type: ignore
    info.Iq = getattr(kernel_module, 'Iq', None) # type: ignore
    info.Iqxy = getattr(kernel_module, 'Iqxy', None) # type: ignore
    info.Iqac = getattr(kernel_module, 'Iqac', None) # type: ignore
    info.Iqabc = getattr(kernel_module, 'Iqabc', None) # type: ignore
    info.Imagnetic = getattr(kernel_module, 'Imagnetic', None) # type: ignore
    info.profile = getattr(kernel_module, 'profile', None) # type: ignore
    info.sesans = getattr(kernel_module, 'sesans', None) # type: ignore
    # Default single and opencl to True for C models.  Python models have callable Iq.
    info.opencl = getattr(kernel_module, 'opencl', not callable(info.Iq))
    info.single = getattr(kernel_module, 'single', not callable(info.Iq))
    info.random = getattr(kernel_module, 'random', None)
    info.hidden = getattr(kernel_module, 'hidden', None) # type: ignore

    # Set control flag for explicitly set parameters, e.g., in the RPA model.
    control = getattr(kernel_module, 'control', None)
    if control is not None:
        parameters[control].is_control = True

    if callable(info.Iq) and parameters.has_2d:
        raise ValueError("oriented python models not supported")

    # CRUFT: support old-style ER() for effective radius
    # radius_effective(mode, ...), if it is present, should be a callable
    # python function accepting the shape parameters, and have a corresponding
    # Iq function in python (Fq in python is not suported at this time).
    # If ER() is present, it might represent a 4.2 model that hasn't been
    # converted, or it might represent a 4.2+5.0 model with both ER and
    # radius_effective defined. Unfortunately we can't easily tell whether
    # radius_effective is defined in the C sources without loading them,
    # so just issue a warning if we see ER in a C model.
    ER = getattr(kernel_module, 'ER', None)
    if ER is not None:
        if callable(info.Iq) and info.radius_effective is None:
            info.radius_effective_modes = ['ER']
            info.radius_effective = lambda mode, *args: ER(*args)
            # TODO: uncomment the following for the sasview 4.3 release
            #logger.warning(
            #    "ER(...) deprecated. Replace with radius_effective(mode, ...)."
            #    " See plugin model docs for details."
            #)
        else:
            logger.warning(
                "ER(...) function ignored. Using radius_effective(mode, ...)"
                " instead if it exists.")
    ## --- untested, so left unsupported for now ---
    ## CRUFT: support old-style VR for form/shell ratio
    #if info.shell_volume is None:
    #    VR = getattr(kernel_module, 'VR')
    #    if VR is not None:
    #        info.shell_volume = lambda *args: info.form_volume(*args)*VR(*args)

    info.lineno = {}
    _find_source_lines(info, kernel_module)
    return info

class ModelInfo(object):
    """
    Interpret the model definition file, categorizing the parameters.

    The module can be loaded with a normal python import statement if you
    know which module you need, or with __import__('sasmodels.model.'+name)
    if the name is in a string.

    The structure should be mostly static, other than the delayed definition
    of *Iq*, *Iqac* and *Iqabc* if they need to be defined.
    """
    #: Full path to the file defining the kernel, if any.
    filename = None         # type: Optional[str]
    #: Base file is usually filename, but not when a model has been
    #: reparameterized, in which case it is the file containing the original
    #: model definition.  This is needed to signal an additional dependency for
    #: the model time stamp, and so that the compiler reports correct file
    #: for syntax errors.
    basefile = None         # type: Optional[str]
    #: Id of the kernel used to load it from the filesystem.
    id = None               # type: str
    #: Display name of the model, which defaults to the model id but with
    #: capitalization of the parts so for example core_shell defaults to
    #: "Core Shell".
    name = None             # type: str
    #: Short description of the model.
    title = None            # type: str
    #: Long description of the model.
    description = None      # type: str
    #: Model parameter table. Parameters are defined using a list of parameter
    #: definitions, each of which is contains parameter name, units,
    #: default value, limits, type and description.  See :class:`Parameter`
    #: for details on the individual parameters.  The parameters are gathered
    #: into a :class:`ParameterTable`, which provides various views into the
    #: parameter list.
    parameters = None       # type: ParameterTable
    #: For reparameterized systems, *base* is the base parameter table.  For
    #: normal systems it is simply a copy of *parameters*.
    base = None             # type: ParameterTable
    #: Parameter translation code to convert from *parameters* table from
    #: caller to the *base* table used to evaluate the model.
    translation = None      # type: Optional[str]
    #: Composition is None if this is an independent model, or it is a
    #: tuple with comoposition type ('product' or 'misture') and a list of
    #: :class:`ModelInfo` blocks for the composed objects.  This allows us
    #: to rebuild a complete mixture or product model from the info block.
    #: *composition* is not given in the model definition file, but instead
    #: arises when the model is constructed using names such as
    #: *sphere*hardsphere* or *cylinder+sphere*.
    composition = None      # type: Optional[Tuple[str, List[ModelInfo]]]
    #: Different variants require different parameters.  In order to show
    #: just the parameters needed for the variant selected,
    #: you should provide a function *hidden(control) -> set(['a', 'b', ...])*
    #: indicating which parameters need to be hidden.  For multiplicity
    #: models, you need to use the complete name of the parameter, including
    #: its number.  So for example, if variant "a" uses only *sld1* and *sld2*,
    #: then *sld3*, *sld4* and *sld5* of multiplicity parameter *sld[5]*
    #: should be in the hidden set.
    hidden = None           # type: Optional[Callable[[int], Set[str]]]
    #: Doc string from the top of the model file.  This should be formatted
    #: using ReStructuredText format, with latex markup in ".. math"
    #: environments, or in dollar signs.  This will be automatically
    #: extracted to a .rst file by :func:`.generate.make_doc`, then
    #: converted to HTML or PDF by Sphinx.
    docs = None             # type: str
    #: Location of the model description in the documentation.  This takes the
    #: form of "section" or "section:subsection".  So for example,
    #: :ref:`porod` uses *category="shape-independent"* so it is in the
    #: :ref:`shape-independent` section whereas
    #: :ref:`capped-cylinder` uses: *category="shape:cylinder"*, which puts
    #: it in the :ref:`shape-cylinder` section.
    category = None         # type: Optional[str]
    #: True if the model can be computed accurately with single precision.
    #: This is True by default, but models such as :ref:`bcc-paracrystal` set
    #: it to False because they require double precision calculations.
    single = None           # type: bool
    #: True if the model can be run as an opencl model.  If for some reason
    #: the model cannot be run in opencl (e.g., because the model passes
    #: functions by reference), then set this to false.
    opencl = None           # type: bool
    #: True if the model is a structure factor used to model the interaction
    #: between form factor models.  This will default to False if it is not
    #: provided in the file.
    structure_factor = None # type: bool
    #: True if the model defines an Fq function with signature
    #: ``void Fq(double q, double *F1, double *F2, ...)``
    have_Fq = False
    #: List of options for computing the effective radius of the shape,
    #: or None if the model is not usable as a form factor model.
    radius_effective_modes = None   # type: List[str]
    #: List of C source files used to define the model.  The source files
    #: should define the *Iq* function, and possibly *Iqac* or *Iqabc* if the
    #: model defines orientation parameters. Files containing the most basic
    #: functions must appear first in the list, followed by the files that
    #: use those functions.
    source = None           # type: List[str]
    #: inline source code, added after all elements of source
    c_code = None           # type: Optional[str]
    #: Expression which evaluates to True if the input parameters are valid
    #: and the model can be computed, or False otherwise.  Invalid parameter
    #: sets will not be included in the weighted $I(Q)$ calculation or its
    #: volume normalization.  Use C syntax for the expressions, with || for or
    #: && for and and ! for not.  Any non-magnetic parameter can be used.
    valid = None         # type: str
    #: Returns the form volume for python-based models.  Form volume is needed
    #: for volume normalization in the polydispersity integral.  If no
    #: parameters are *volume* parameters, then form volume is not needed.
    #: For C-based models, (with :attr:`source` defined, or with :attr:`Iq`
    #: defined using a string containing C code), form_volume must also be
    #: C code, either defined as a string, or in the sources.
    form_volume = None      # type: Union[None, str, Callable[[np.ndarray], float]]
    #: Returns the shell volume for python-based models.  Form volume and
    #: shell volume are needed for volume normalization in the polydispersity
    #: integral and structure interactions for hollow shapes.  If no
    #: parameters are *volume* parameters, then shell volume is not needed.
    #: For C-based models, (with :attr:`source` defined, or with :attr:`Iq`
    #: defined using a string containing C code), shell_volume must also be
    #: C code, either defined as a string, or in the sources.
    shell_volume = None      # type: Union[None, str, Callable[[np.ndarray], float]]
    #: Computes the effective radius of the shape given the volume parameters.
    #: Only needed for models defined in python that can be used for
    #: monodisperse approximation for non-dilute solutions, P@S.  The first
    #: argument is the integer effective radius mode, with default 0.
    radius_effective = None  # type: Union[None, Callable[[int, np.ndarray], float]]
    #: Returns *I(q, a, b, ...)* for parameters *a*, *b*, etc. defined
    #: by the parameter table.  *Iq* can be defined as a python function, or
    #: as a C function.  If it is defined in C, then set *Iq* to the body of
    #: the C function, including the return statement.  This function takes
    #: values for *q* and each of the parameters as separate *double* values
    #: (which may be converted to float or long double by sasmodels).  All
    #: source code files listed in :attr:`source` will be loaded before the
    #: *Iq* function is defined.  If *Iq* is not present, then sources should
    #: define *static double Iq(double q, double a, double b, ...)* which
    #: will return *I(q, a, b, ...)*.  Multiplicity parameters are sent as
    #: pointers to doubles.  Constants in floating point expressions should
    #: include the decimal point. See :mod:`.generate` for more details. If
    #: *have_Fq* is True, then Iq should return an interleaved array of
    #: $[\sum F(q_1), \sum F^2(q_1), \ldots, \sum F(q_n), \sum F^2(q_n)]$.
    Iq = None               # type: Union[None, str, Callable[[...], np.ndarray]]
    #: Returns *I(qx, qy, a, b, ...)*.  The interface follows :attr:`Iq`.
    Iqxy = None             # type: Union[None, str, Callable[[...], np.ndarray]]
    #: Returns *I(qab, qc, a, b, ...)*.  The interface follows :attr:`Iq`.
    Iqac = None             # type: Union[None, str, Callable[[...], np.ndarray]]
    #: Returns *I(qa, qb, qc, a, b, ...)*.  The interface follows :attr:`Iq`.
    Iqabc = None            # type: Union[None, str, Callable[[...], np.ndarray]]
    #: Returns *I(qx, qy, a, b, ...)*.  The interface follows :attr:`Iq`.
    Imagnetic = None        # type: Union[None, str, Callable[[...], np.ndarray]]
    #: Returns a model profile curve *x, y*.  If *profile* is defined, this
    #: curve will appear in response to the *Show* button in SasView.  Use
    #: :attr:`profile_axes` to set the axis labels.  Note that *y* values
    #: will be scaled by 1e6 before plotting.
    profile = None          # type: Optional[Callable[[np.ndarray], None]]
    #: Axis labels for the :attr:`profile` plot.  The default is *['x', 'y']*.
    #: Only the *x* component is used for now.
    profile_axes = None     # type: Tuple[str, str]
    #: Returns *sesans(z, a, b, ...)* for models which can directly compute
    #: the SESANS correlation function.  Note: not currently implemented.
    sesans = None           # type: Optional[Callable[[np.ndarray], np.ndarray]]
    #: Returns a random parameter set for the model
    random = None           # type: Optional[Callable[[], Dict[str, float]]]
    #: Line numbers for symbols defining C code
    lineno = None           # type: Dict[str, int]
    #: The set of tests that must pass.  The format of the tests is described
    #: in :mod:`.model_test`.
    tests = None            # type: List[TestCondition]

    def __init__(self):
        # type: () -> None
        pass

    def get_hidden_parameters(self, control):
        """
        Returns the set of hidden parameters for the model.  *control* is the
        value of the control parameter.  Note that multiplicity models have
        an implicit control parameter, which is the parameter that controls
        the multiplicity.
        """
        if self.hidden is not None:
            hidden = self.hidden(control)
        else:
            controls = [p for p in self.parameters.kernel_parameters
                        if p.is_control]
            if len(controls) != 1:
                raise ValueError("more than one control parameter")
            hidden = set(p.id+str(k)
                         for p in self.parameters.kernel_parameters
                         for k in range(control+1, p.length+1)
                         if p.length > 1)
            for p in self.parameters.kernel_parameters:
                if p.length > 1 and p.type == "sld":
                    for k in range(control+1, p.length+1):
                        base = p.id+str(k)
                        hidden.update((base+"_M0", base+"_mtheta", base+"_mphi"))
        return hidden



# -*- coding: utf-8 -*-
"""
Run model unit tests.

Usage::

    python -m sasmodels.model_test [opencl|cuda|dll|all] model1 model2 ...

If model1 is 'all', then all except the remaining models will be tested.
Subgroups are also possible, such as 'py', 'single' or '1d'.  See
:func:`.core.list_models` for details.

Each model is tested using the default parameters at q=0.1, (qx, qy)=(0.1, 0.1),
and Fq is called to make sure R_eff, volume and volume ratio are computed.
The return values at these points are not considered.  The test is only to
verify that the models run to completion, and do not produce inf or NaN.

Tests are defined with the *tests* attribute in the model.py file.  *tests*
is a list of individual tests to run, where each test consists of the
parameter values for the test, the q-values and the expected results.  For
the effective radius test and volume ratio tests, use the extended output
form, which checks each output of kernel.Fq. For 1-D tests, either specify
the q value or a list of q-values, and the corresponding I(q) value, or
list of I(q) values.

That is::

    tests = [
        [ {parameters}, q, I(q)],
        [ {parameters}, [q], [I(q)] ],
        [ {parameters}, [q1, q2, ...], [I(q1), I(q2), ...]],

        [ {parameters}, (qx, qy), I(qx, Iqy)],
        [ {parameters}, [(qx1, qy1), (qx2, qy2), ...],
                        [I(qx1, qy1), I(qx2, qy2), ...]],

        [ {parameters}, q, F(q), F^2(q), R_eff, V, V_r ],
        ...
    ]

Parameters are *key:value* pairs, where key is one of the parameters of the
model and value is the value to use for the test.  Any parameters not given
in the parameter list will take on the default parameter value.

Precision defaults to 5 digits (relative).
"""
from __future__ import print_function

import argparse
import sys
import unittest
import traceback

try:
    from StringIO import StringIO
except ImportError:
    # StringIO.StringIO renamed to io.StringIO in Python 3
    # Note: io.StringIO exists in python 2, but using unicode instead of str
    from io import StringIO

import numpy as np  # type: ignore

from .core import list_models, load_model_info, build_model
from .direct_model import call_kernel, call_Fq
from .exception import annotate_exception
from .modelinfo import expand_pars
from .kernelcl import use_opencl
from .kernelcuda import use_cuda
from . import product

# pylint: disable=unused-import
try:
    from typing import List, Iterator, Callable, Any, Dict, Tuple, Union
    from .modelinfo import ParameterTable, ParameterSet, TestCondition, ModelInfo
    from .kernel import KernelModel
    DType = Union[None, str, np.dtype]
except ImportError:
    pass
# pylint: enable=unused-import

def make_suite(loaders, models):
    # type: (List[str], List[str]) -> unittest.TestSuite
    """
    Construct the pyunit test suite.

    *loaders* is the list of kernel drivers to use (dll, opencl or cuda).
    For python model the python driver is always used.

    *models* is the list of models to test, or *["all"]* to test all models.
    """
    suite = unittest.TestSuite()

    try:
        # See if the first model parses as a model group
        group = list_models(models[0])
        skip = models[1:]
        models = group
    except Exception:
        skip = []
    for model_name in models:
        if model_name not in skip:
            model_info = load_model_info(model_name)
            _add_model_to_suite(loaders, suite, model_info)

    return suite

def _add_model_to_suite(loaders, suite, model_info):
    ModelTestCase = _hide_model_case_from_nose()

    #print('------')
    #print('found tests in', model_name)
    #print('------')

    # if ispy then use the dll loader to call pykernel
    # don't try to call cl kernel since it will not be
    # available in some environmentes.
    is_py = callable(model_info.Iq)

    # Some OpenCL drivers seem to be flaky, and are not producing the
    # expected result.  Since we don't have known test values yet for
    # all of our models, we are instead going to compare the results
    # for the 'smoke test' (that is, evaluation at q=0.1 for the default
    # parameters just to see that the model runs to completion) between
    # the OpenCL and the DLL.  To do this, we define a 'stash' which is
    # shared between OpenCL and DLL tests.  This is just a list.  If the
    # list is empty (which it will be when DLL runs, if the DLL runs
    # first), then the results are appended to the list.  If the list
    # is not empty (which it will be when OpenCL runs second), the results
    # are compared to the results stored in the first element of the list.
    # This is a horrible stateful hack which only makes sense because the
    # test suite is thrown away after being run once.
    stash = []

    if is_py:  # kernel implemented in python
        test_name = "%s[python]"%model_info.name
        test_method_name = "test_%s_python" % model_info.id
        test = ModelTestCase(test_name, model_info,
                             test_method_name,
                             platform="dll",  # so that
                             dtype="double",
                             stash=stash)
        suite.addTest(test)
    else:   # kernel implemented in C

        # test using dll if desired
        if 'dll' in loaders or not use_opencl():
            test_name = "%s[dll]"%model_info.name
            test_method_name = "test_%s_dll" % model_info.id
            test = ModelTestCase(test_name, model_info,
                                 test_method_name,
                                 platform="dll",
                                 dtype="double",
                                 stash=stash)
            suite.addTest(test)

        # test using opencl if desired and available
        if 'opencl' in loaders and use_opencl():
            test_name = "%s[opencl]"%model_info.name
            test_method_name = "test_%s_opencl" % model_info.id
            # Using dtype=None so that the models that are only
            # correct for double precision are not tested using
            # single precision.  The choice is determined by the
            # presence of *single=False* in the model file.
            test = ModelTestCase(test_name, model_info,
                                 test_method_name,
                                 platform="ocl", dtype=None,
                                 stash=stash)
            #print("defining", test_name)
            suite.addTest(test)

        # test using cuda if desired and available
        if 'cuda' in loaders and use_cuda():
            test_name = "%s[cuda]" % model_info.id
            test_method_name = "test_%s_cuda" % model_info.id
            # Using dtype=None so that the models that are only
            # correct for double precision are not tested using
            # single precision.  The choice is determined by the
            # presence of *single=False* in the model file.
            test = ModelTestCase(test_name, model_info,
                                 test_method_name,
                                 platform="cuda", dtype=None,
                                 stash=stash)
            #print("defining", test_name)
            suite.addTest(test)


def _hide_model_case_from_nose():
    # type: () -> type
    class ModelTestCase(unittest.TestCase):
        """
        Test suit for a particular model with a particular kernel driver.

        The test suite runs a simple smoke test to make sure the model
        functions, then runs the list of tests at the bottom of the model
        description file.
        """
        def __init__(self, test_name, model_info, test_method_name,
                     platform, dtype, stash):
            # type: (str, ModelInfo, str, str, DType, List[Any]) -> None
            self.test_name = test_name
            self.info = model_info
            self.platform = platform
            self.dtype = dtype
            self.stash = stash  # container for the results of the first run
            self._failures = []  # Set of failed target values

            setattr(self, test_method_name, self.run_all)
            unittest.TestCase.__init__(self, test_method_name)

        def run_all(self):
            # type: () -> None
            """
            Run all the tests in the test suite, including smoke tests.
            """
            smoke_tests = [
                # test validity at reasonable values
                ({}, 0.1, None),
                ({}, (0.1, 0.1), None),
                # test validity at q = 0
                #({}, 0.0, None),
                #({}, (0.0, 0.0), None),
                # test vector form
                ({}, [0.001, 0.01, 0.1], [None]*3),
                ({}, [(0.1, 0.1)]*2, [None]*2),
                # test that Fq will run, and return R_eff, V, V_r
                ({}, 0.1, None, None, None, None, None),
                ]
            tests = smoke_tests
            #tests = []
            if self.info.tests is not None:
                tests += self.info.tests
            S_tests = [test for test in tests if '@S' in test[0]]
            P_tests = [test for test in tests if '@S' not in test[0]]
            del self._failures[:]
            try:
                model = build_model(self.info, dtype=self.dtype,
                                    platform=self.platform)
                results = [self.run_one(model, test) for test in P_tests]
                for test in S_tests:
                    # pull the S model name out of the test defn
                    pars = test[0].copy()
                    s_name = pars.pop('@S')
                    ps_test = [pars] + list(test[1:])
                    #print("PS TEST PARAMS!!!",ps_test)
                    # build the P@S model
                    s_info = load_model_info(s_name)
                    ps_info = product.make_product_info(self.info, s_info)
                    ps_model = build_model(ps_info, dtype=self.dtype,
                                           platform=self.platform)
                    # run the tests
                    #self.info = ps_model.info
                    #print("SELF.INFO PARAMS!!!",[p.id for p in self.info.parameters.call_parameters])
                    #print("PS MODEL PARAMETERS:",[p.id for p in ps_model.info.parameters.call_parameters])
                    results.append(self.run_one(ps_model, ps_test))

                if self.stash:
                    for test, target, actual in zip(tests, self.stash[0], results):
                        assert np.all(abs(target-actual) < 5e-5*abs(actual)), \
                            ("GPU/CPU comparison expected %s but got %s for %s"
                             % (target, actual, test[0]))
                else:
                    self.stash.append(results)

                # Check for missing tests.  Only do so for the "dll" tests
                # to reduce noise from both opencl and cuda, and because
                # python kernels use platform="dll".
                if self.platform == "dll":
                    missing = []
                    ## Uncomment the following to require test cases
                    #missing = self._find_missing_tests()
                    if missing:
                        raise ValueError("Missing tests for "+", ".join(missing))

            except:
                annotate_exception(self.test_name)
                raise

            if self._failures:
                msg = ("The following assertions failed for %s:\n  %s"
                       % (self.test_name, "\n  ".join(self._failures)))
                raise AssertionError(msg)

        def _find_missing_tests(self):
            # type: () -> None
            """make sure there are 1D and 2D tests as appropriate"""
            model_has_1D = True
            model_has_2D = any(p.type == 'orientation'
                               for p in self.info.parameters.kernel_parameters)

            # Lists of tests that have a result that is not None
            single = [test for test in self.info.tests
                      if not isinstance(test[2], list) and test[2] is not None]
            tests_has_1D_single = any(isinstance(test[1], float) for test in single)
            tests_has_2D_single = any(isinstance(test[1], tuple) for test in single)

            multiple = [test for test in self.info.tests
                        if isinstance(test[2], list)
                        and not all(result is None for result in test[2])]
            tests_has_1D_multiple = any(isinstance(test[1][0], float)
                                        for test in multiple)
            tests_has_2D_multiple = any(isinstance(test[1][0], tuple)
                                        for test in multiple)

            missing = []
            if model_has_1D and not (tests_has_1D_single or tests_has_1D_multiple):
                missing.append("1D")
            if model_has_2D and not (tests_has_2D_single or tests_has_2D_multiple):
                missing.append("2D")

            return missing

        def run_one(self, model, test):
            # type: (KernelModel, TestCondition) -> None
            """Run a single test case."""
            user_pars, x, y = test[:3]
            #print("PS MODEL PARAMETERS:",[p.id for p in model.info.parameters.call_parameters])
            pars = expand_pars(model.info.parameters, user_pars)
            invalid = invalid_pars(model.info.parameters, pars)
            if invalid:
                raise ValueError("Unknown parameters in test: " + ", ".join(invalid))

            if not isinstance(y, list):
                y = [y]
            if not isinstance(x, list):
                x = [x]

            self.assertEqual(len(y), len(x))

            if isinstance(x[0], tuple):
                qx, qy = zip(*x)
                q_vectors = [np.array(qx), np.array(qy)]
            else:
                q_vectors = [np.array(x)]

            kernel = model.make_kernel(q_vectors)
            if len(test) == 3 or len(test) == 4:
                actual = call_kernel(kernel, pars)
                self._check_vectors(x, y, actual, 'I')
                if len(test) == 4:
                    results = getattr(kernel, 'results', lambda: {})
                    self._check_struct(x, test[3], results())
                return actual
            elif len(test) == 7:
                y1 = y
                y2 = test[3] if isinstance(test[3], list) else [test[3]]
                F, Fsq, R_eff, volume, volume_ratio = call_Fq(kernel, pars)
                if F is not None:  # F is none for models with Iq instead of Fq
                    self._check_vectors(x, y1, F, 'F')
                self._check_vectors(x, y2, Fsq, 'F^2')
                self._check_scalar(test[4], R_eff, 'R_eff')
                self._check_scalar(test[5], volume, 'volume')
                self._check_scalar(test[6], volume_ratio, 'form:shell ratio')
                return Fsq
            else:
                self._failures.append('wrong number or results for %s => %s'
                                      % (str(user_pars), str(test[2:])))
                return None

        def _check_scalar(self, target, actual, name):
            if not is_near(target, actual, 5):
                self._failures.append('%s: expected:%s; actual:%s'
                                      % (name, target, actual))

        def _check_vectors(self, x, target, actual, name='I'):
            if not len(actual) > 0:
                self._failures.append('%s(...) expected return value'%name)
                return
            if target is None:
                return
            if len(target) != len(actual):
                self._failures.append('%s(...) returned wrong length in %s'
                                      % (name, str(actual)))
                return
            for xi, yi, actual_yi in zip(x, target, actual):
                if not is_near(yi, actual_yi, 5):
                    # convert array to list so we have comma separated output
                    actual = actual.tolist()
                    self._failures.append('%s(%s): expected:%s; actual:%s'
                                          % (name, xi, target, actual))
                    # Whole list is printed on any error, so fail on first
                    return

        def _check_struct(self, x, target, actual):
            for k, v in target.items():
                if k not in actual:
                    self._failures.append('key %r not in returned value' % k)
                    continue
                target_k = v
                actual_k = actual[k]
                if np.isscalar(actual_k):  # number
                    self._check_scalar(target_k, actual_k, k)
                elif isinstance(actual_k, np.ndarray): # vector
                    self._check_vectors(x, target_k, actual_k, k)
                elif isinstance(actual_k, tuple) and len(actual_k) == 2:
                    # Intermediate is returned as (Q, I(Q)) pair.  The test
                    # is set up to use |Q| or (Qx, Qy) as input, but the
                    # returned result has (|Q|,) or (Qx, Qy).  For now we will
                    # ignore the first result.
                    #self._check_vectors(x, x, actual_k[0], k+" (Q)")
                    self._check_vectors(x, target_k, actual_k[1], k)
                elif isinstance(actual_k, dict):
                    self._check_struct(x, target_k, actual_k)
                else:
                    self._failures.append('key %s: unexpected value %r'
                                          % (k, actual_k))
            # Ignore effective_radius in returned result; it will be dropped
            actual.pop('effective_radius', None)
            if any(k not in target for k in actual):
                self._failures.append('intermediate results missing:')
                for k, v in actual.items():
                    if k not in target:
                        if isinstance(v, tuple) and len(v) == 2:
                            v = v[1].tolist()
                        elif isinstance(v, np.ndarray):
                            v = v.tolist()
                        self._failures.append('  "%s": %r,' % (k, v))

    return ModelTestCase

def invalid_pars(partable, pars):
    # type: (ParameterTable, Dict[str, float]) -> List[str]
    """
    Return a list of parameter names that are not part of the model.
    """
    names = set(p.id for p in partable.call_parameters)
    invalid = []
    for par in sorted(pars.keys()):
        # Ignore the R_eff mode parameter when checking for valid parameters.
        # It is an allowed parameter for a model even though it does not exist
        # in the parameter table.  The call_Fq() function pops it from the
        # parameter list and sends it directly to kernel.Fq().
        if par == product.RADIUS_MODE_ID:
            continue
        parts = par.split('_pd')
        if len(parts) > 1 and parts[1] not in ("", "_n", "nsigma", "type"):
            invalid.append(par)
            continue
        if parts[0] not in names:
            invalid.append(par)
    return invalid


def is_near(target, actual, digits=5):
    # type: (float, float, int) -> bool
    """
    Returns true if *actual* is within *digits* significant digits of *target*.

    *taget* zero and inf should match *actual* zero and inf.  If you want to
    accept eps for zero, choose a value such as 1e-10, which must match up to
    +/- 1e-15 when *digits* is the default value of 5.

    If *target* is None, then just make sure that *actual* is not NaN.

    If *target* is NaN, make sure *actual* is NaN.
    """
    if target is None:
        # target is None => actual cannot be NaN
        return not np.isnan(actual)
    elif target == 0.:
        # target is 0. => actual must be 0.
        # Note: if small values are allowed, then use maybe test zero against eps instead?
        return actual == 0.
    elif np.isfinite(target):
        shift = np.ceil(np.log10(abs(target)))
        return abs(target-actual) < 1.5*10**(shift-digits)
    elif target == actual:
        # target is inf => actual must be inf of same sign
        return True
    else:
        # target is NaN => actual must be NaN
        return np.isnan(target) == np.isnan(actual)

# CRUFT: old interface; should be deprecated and removed
def run_one(model_name):
    # type: (str) -> str
    """
    [Deprecated] Run the tests associated with *model_name*.

    Use the following instead::

        succss, output = check_model(load_model_info(model_name))
    """
    # msg = "use check_model(model_info) rather than run_one(model_name)"
    # warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    try:
        model_info = load_model_info(model_name)
    except Exception:
        output = traceback.format_exc()
        return output

    _, output = check_model(model_info)
    return output

def check_model(model_info):
    # type: (ModelInfo) -> Tuple[bool, str]
    """
    Run the tests for a single model, capturing the output.

    Returns success status and the output string.
    """
    # Note that running main() directly did not work from within the
    # wxPython pycrust console.  Instead of the results appearing in the
    # window they were printed to the underlying console.
    from unittest.runner import TextTestResult, _WritelnDecorator

    # Build a object to capture and print the test results
    stream = _WritelnDecorator(StringIO())  # Add writeln() method to stream
    verbosity = 2
    descriptions = True
    result = TextTestResult(stream, descriptions, verbosity)

    # Build a test suite containing just the model
    loaders = ['opencl' if use_opencl() else 'cuda' if use_cuda() else 'dll']
    suite = unittest.TestSuite()
    _add_model_to_suite(loaders, suite, model_info)

    # Warn if there are no user defined tests.
    # Note: the test suite constructed above only has one test in it, which
    # runs through some smoke tests to make sure the model runs, then runs
    # through the input-output pairs given in the model definition file.  To
    # check if any such pairs are defined, therefore, we just need to check if
    # they are in the first test of the test suite.  We do this with an
    # iterator since we don't have direct access to the list of tests in the
    # test suite.
    # In Qt5 suite.run() will clear all tests in the suite after running
    # with no way of retaining them for the test below, so let's check
    # for user tests before running the suite.
    for test in suite:
        if not test.info.tests:
            stream.writeln("Note: %s has no user defined tests."%model_info.name)
        break
    else:
        stream.writeln("Note: no test suite created --- this should never happen")

    # Run the test suite
    suite.run(result)

    # Print the failures and errors
    for _, tb in result.errors:
        stream.writeln(tb)
    for _, tb in result.failures:
        stream.writeln(tb)

    output = stream.getvalue()
    stream.close()
    return result.wasSuccessful(), output


def model_tests():
    # type: () -> Iterator[Callable[[], None]]
    """
    Test runner visible to nosetests.

    Run "nosetests sasmodels" on the command line to invoke it.
    """
    loaders = ['dll']
    if use_opencl():
        loaders.append('opencl')
    if use_cuda():
        loaders.append('cuda')
    tests = make_suite(loaders, ['all'])
    def _build_test(test):
        # In order for nosetest to show the test name, wrap the test.run_all
        # instance in function that takes the test name as a parameter which
        # will be displayed when the test is run.  Do this as a function so
        # that it properly captures the context for tests that captured and
        # run later.  If done directly in the for loop, then the looping
        # variable test will be shared amongst all the tests, and we will be
        # repeatedly testing vesicle.

        # Note: in sasview sas.sasgui.perspectives.fitting.gpu_options
        # requires that the test.description field be set.
        wrap = lambda: test.run_all()
        wrap.description = test.test_name
        return wrap
        # The following would work with nosetests and pytest:
        #     return lambda name: test.run_all(), test.test_name

    for test in tests:
        yield _build_test(test)


def main():
    # type: () -> int
    """
    Run tests given is models.

    Returns 0 if success or 1 if any tests fail.
    """
    try:
        from xmlrunner import XMLTestRunner as TestRunner
        test_args = {'output': 'logs'}
    except ImportError:
        from unittest import TextTestRunner as TestRunner
        test_args = {}

    parser = argparse.ArgumentParser(description="Test SasModels Models")
    parser.add_argument("-v", "--verbose", action="store_const",
                        default=1, const=2, help="Use verbose output")
    parser.add_argument("-e", "--engine", default="all",
                        help="Engines on which to run the test.  "
                        "Valid values are opencl, cuda, dll, and all. "
                        "Defaults to all if no value is given")
    parser.add_argument("models", nargs="*",
                        help="The names of the models to be tested.  "
                        "If the first model is 'all', then all but the listed "
                        "models will be tested.  See core.list_models() for "
                        "names of other groups, such as 'py' or 'single'.")
    opts = parser.parse_args()

    if opts.engine == "opencl":
        if not use_opencl():
            print("opencl is not available")
            return 1
        loaders = ['opencl']
    elif opts.engine == "dll":
        loaders = ["dll"]
    elif opts.engine == "cuda":
        if not use_cuda():
            print("cuda is not available")
            return 1
        loaders = ['cuda']
    elif opts.engine == "all":
        loaders = ['dll']
        if use_opencl():
            loaders.append('opencl')
        if use_cuda():
            loaders.append('cuda')
    else:
        print("unknown engine " + opts.engine)
        return 1

    runner = TestRunner(verbosity=opts.verbose, **test_args)
    result = runner.run(make_suite(loaders, opts.models))
    return 1 if result.failures or result.errors else 0


if __name__ == "__main__":
    sys.exit(main())



r"""
Definition
----------

Calculates the scattering from a **body-centered cubic lattice** with
paracrystalline distortion. Thermal vibrations are considered to be negligible,
and the size of the paracrystal is infinitely large. Paracrystalline distortion
is assumed to be isotropic and characterized by a Gaussian distribution.

The scattering intensity $I(q)$ is calculated as

.. math::

    I(q) = \frac{\text{scale}}{V_p} V_\text{lattice} P(q) Z(q) + \text{background}

where *scale* is the volume fraction of crystal in the sample volume,
$V_\text{lattice}$ is the volume fraction of spheres in the crystal, $V_p$ is
the volume of the primary particle, $P(q)$ is the form factor of the sphere
(normalized), and $Z(q)$ is the paracrystalline structure factor for a
body-centered cubic structure.

.. note::
    At this point the GUI does not return $V_\text{lattice}$ separately so that
    the user will need to calculate it from the equation given and the
    appropriate returned parameters.

.. warning::
    As per the equations below, this model will return I(q)=0 for all q if the
    distortion factor is equal to 0. The model is not meant to support perfect
    crystals.

.. figure:: img/bcc_geometry.jpg

    Body-centered cubic (BCC) lattice taken from reference [#Matsuoka1987]_.

Following the derivation from reference [#Matsuoka1987]_, as corrected in
reference [#Matsuoka1990]_, and based on the above figure, the
primitive unit cell vectors $\vec{a_1},\vec{a_2}$, and $\vec{a_3}$, which
enclose the smallest possible unit cell for the bcc lattice, are defined below:

.. math::
    \vec{a_1} &= \frac{1}{2}(-\vec{b_1} + \vec{b_2} + \vec{b_3}) \\
    \vec{a_2} &= \frac{1}{2} (\vec{b_1} - \vec{b_2} + \vec{b_3}) \\
    \vec{a_3} &= \frac{1}{2}(\vec{b_1} + \vec{b_2} -\vec{b_3}).

where $\vec{b_1},\vec{b_2}$, and $\vec{b_3}$ are the unit cell vectors of the
conventional unit cell, which is a unit cell that includes the full symmetry
of the lattice. As defined by reference [#Matsuoka1987]_, the constant $a$ is the
lattice parameter of the conventional unit cell with
$|\vec{b_1}|=|\vec{b_2}|=|\vec{b_3}|=a$. Using this definition, the
nearest-neighbor distance ($D$) is given by
$D=|\vec{a_1}|=|\vec{a_2}|=|\vec{a_3}|=\sqrt{(a/2)^2+(a/2)^2+(a/2)^2}=\sqrt{\frac{3a^2}{4}}=\frac{\sqrt{3}a}{2}$.

The volume of the primitive unit cell $V_u$ is then given by:

.. math::
    V_u &= |(\vec{a_1}\times \vec{a_2})\cdot\vec{a_3}|\\
    &= (\frac{a^2}{2},\frac{a^2}{2},0)\cdot(\frac{a}{2},\frac{a}{2},-\frac{a}{2})\\
    &= a^3/2

In this case, the volume fraction ($V_{lattice}$) of spherical particles with
radius $R$ sitting on the bcc lattice is given by:

.. math::
    V_{lattice} &= \frac{4/3 \pi R^3}{a^3/2}\\
    &= \frac{8\pi R^3}{3a^3}\\
    &= \frac{\sqrt{3} \pi R^3}{D^3}

Now, continuing to follow [#Matsuoka1987]_, the structure (lattice)
factor $Z(\vec{q})$ for a 3D paracrystal can be written as:

.. math::
    Z(\vec{q}) = \prod_{k=1}^{3}Z_k(\vec{q})

with

.. math::
    Z_k(\vec{q}) = \frac{1-|F_k|^2}{1-2|F_k|\cos(\vec{a_k}\cdot\vec{q})+|F_k|^2}

and where $F_k(\vec{q})$ is the structure factor of the primitive unit cell
defined as:

.. math::
    F_k(\vec{q}) = e^{-\frac{1}{2} \Delta a^2_k q^2} \times e^{-i\vec{q}\cdot\vec{a_k}}.

Here, $\vec{a_k}$ are the primitive unit cell vectors $\vec{a_1}$, $\vec{a_2}$,
and $\vec{a_3}$. Furthermore, $\Delta a_k$ is the isotropic distortion of the
lattice point from its ideal position and can be defined by a constant factor
$g=\Delta a / |\vec{a_1}| = \Delta a / |\vec{a_2}| = \Delta a / |\vec{a_3}|=\Delta a/D$.

Finally, assuming the definitions presented in this document, the authors of
reference [#Matsuoka1987]_ have derived the lattice factors which are given by:

.. math::
    Z_1(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qa}{2}(\sin\theta \cos\phi + \sin\theta \sin\phi + \cos\theta)] + e^{-q^2\Delta a^2}\}\\
    Z_2(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qa}{2}(-\sin\theta \cos\phi - \sin\theta \sin\phi + \cos\theta)] + e^{-q^2\Delta a^2}\}\\
    Z_3(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qa}{2}(-\sin\theta \cos\phi + \sin\theta \sin\phi - \cos\theta)] + e^{-q^2\Delta a^2}\}\\

Note that Sasview is using the nearest-neighbor parameter ($D$) as an input
instead of the conventional unit cell parameter $a$. In this case, using
$a=\frac{2D}{\sqrt{3}}$, we rewrite $Z_1(q)$, $Z_2(q)$, and $Z_3(q)$ in terms
of $D$ instead of $a$, which leads to:

.. math::
    Z_1(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qD}{\sqrt{3}}(\sin\theta \cos\phi + \sin\theta \sin\phi + \cos\theta)] + e^{-q^2\Delta a^2}\}\\
    Z_2(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qD}{\sqrt{3}}(-\sin\theta \cos\phi - \sin\theta \sin\phi + \cos\theta)] + e^{-q^2\Delta a^2}\}\\
    Z_3(q,\theta,\phi)&=[1-e^{-q^2\Delta a^2}]/\{1-2e^{-\frac{1}{2}q^2\Delta a^2}\cos[\frac{qD}{\sqrt{3}}(-\sin\theta \cos\phi + \sin\theta \sin\phi - \cos\theta)] + e^{-q^2\Delta a^2}\}\\

Finally note that the position of the Bragg peaks for the bcc lattice are
indexed by (reduced q-values):

.. math::
    \frac{qa}{2\pi}=\frac{qD}{\sqrt{3}\pi}=\sqrt{h^2+k^2+l^2}.

In the above equation, we used the conventional unit cell so not all
permutations of h,k, and l will produce Bragg peaks. The Bragg scattering
condition for bcc imposes that h+k+l = even. Thus the peak positions
correspond to (just the first 5)

.. math::

    \begin{array}{lccccc}
    q/q_o          &   1   & \sqrt{2} & \sqrt{3} & \sqrt{4} & \sqrt{5} \\
    \text{Indices} & (110) &    (200) & (211)    & (220)    & (310)    \\
    \end{array}

.. note::

  The calculation of $Z(q)$ is a double numerical integral that must be
  carried out with a high density of points to properly capture the sharp
  peaks of the paracrystalline scattering. So be warned that the calculation
  is slow. Fitting of any experimental data must be resolution smeared for
  any meaningful fit. This makes a triple integral which may be very slow.
  If a double-precision GPU with OpenCL support is available this may improve
  the speed of the calculation.

This example dataset is produced using 200 data points,
*qmin* = 0.001 |Ang^-1|, *qmax* = 0.1 |Ang^-1| and the above default values.

The 2D (Anisotropic model) is based on the reference below where $I(q)$ is
approximated for 1d scattering. Thus the scattering pattern for 2D may not be
accurate, particularly at low $q$. For general details of the calculation and
angular dispersions for oriented particles see :ref:`orientation`. Note that
we are not responsible for any incorrectness of the 2D model computation.

.. figure:: img/parallelepiped_angle_definition.png

    Orientation of the crystal with respect to the scattering plane, when
    $\theta = \phi = 0$ the $c$ axis is along the beam direction (the $z$ axis).

References
----------

.. [#Matsuoka1987] Hideki Matsuoka et. al. *Physical Review B*, 36 (1987)
   1754-1765 (Original Paper)
.. [#Matsuoka1990] Hideki Matsuoka et. al. *Physical Review B*, 41 (1990)
   3854-3856 (Corrections to FCC and BCC lattice structure calculation)

Authorship and Verification
---------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Jonathan Gaudet **Date:** September 26, 2022
* **Last Reviewed by:** Paul Butler **Date:** November 2, 2022
"""

import numpy as np
from numpy import inf, pi

name = "bcc_paracrystal"
title = "Body-centred cubic lattic with paracrystalline distortion"
description = """
    Calculates the scattering from a **body-centered cubic lattice** with
    paracrystalline distortion. Thermal vibrations are considered to be
    negligible, and the size of the paracrystal is infinitely large.
    Paracrystalline distortion is assumed to be isotropic and characterized
    by a Gaussian distribution.
    """
category = "shape:paracrystal"

#note - calculation requires double precision
single = False

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type","description" ],
parameters = [["dnn",         "Ang",       220,    [-inf, inf], "",            "Nearest neighbour distance"],
              ["d_factor",    "",            0.06, [-inf, inf], "",            "Paracrystal distortion factor"],
              ["radius",      "Ang",        40,    [0, inf],    "volume",      "Particle radius"],
              ["sld",         "1e-6/Ang^2",  4,    [-inf, inf], "sld",         "Particle scattering length density"],
              ["sld_solvent", "1e-6/Ang^2",  1,    [-inf, inf], "sld",         "Solvent scattering length density"],
              ["theta",       "degrees",    60,    [-360, 360], "orientation", "c axis to beam angle"],
              ["phi",         "degrees",    60,    [-360, 360], "orientation", "rotation about beam"],
              ["psi",         "degrees",    60,    [-360, 360], "orientation", "rotation about c axis"]
             ]
# pylint: enable=bad-whitespace, line-too-long

source = ["lib/sas_3j1x_x.c", "lib/gauss150.c", "lib/sphere_form.c", "bcc_paracrystal.c"]

def random():
    """Return a random parameter set for the model."""
    # Define lattice spacing as a multiple of the particle radius
    # using the formula a = 4 r/sqrt(3).  Systems which are ordered
    # are probably mostly filled, so use a distribution which goes from
    # zero to one, but leaving 90% of them within 80% of the
    # maximum bcc packing.  Lattice distortion values are empirically
    # useful between 0.01 and 0.7.  Use an exponential distribution
    # in this range 'cuz its easy.
    radius = 10**np.random.uniform(1.3, 4)
    d_factor = 10**np.random.uniform(-2, -0.7)  # sigma_d in 0.01-0.7
    dnn_fraction = np.random.beta(a=10, b=1)
    dnn = radius*4/np.sqrt(3)/dnn_fraction
    pars = dict(
        #sld=1, sld_solvent=0, scale=1, background=1e-32,
        dnn=dnn,
        d_factor=d_factor,
        radius=radius,
    )
    return pars

# april 6 2017, rkh add unit tests, NOT compared with any other calc method, assume correct!
# add 2d test later

# October 26, 2022 PDB updated the 1D unit test after fixing the math. The values are again
# assumed correct. It would be good to have an independent assessment. 2D tests remain
# on the todo list
# TODO: fix the 2d tests
q = 4.*pi/220.
tests = [
    [{}, [0.001, q, 0.25], [0.6945817843046642, 1.6885157981411993, 0.005367008206852725]],
    #[{'theta': 20.0, 'phi': 30, 'psi': 40.0}, (-0.017, 0.035), 2082.20264399],
    #[{'theta': 20.0, 'phi': 30, 'psi': 40.0}, (-0.081, 0.011), 0.436323144781],
    ]



"""
pytest hooks for sasmodels

Hooks for running sasmodels tests via pytest.

*pytest_collection_modifyitems* adds the test description to the end of
the test name.  This is needed for the generated list of tests is sasmodels,
where each test has a description giving the name of the model.  For example
"model_tests::[3]" becomes "model_tests::[3]::bcc_paracrystal-dll".  Need to
leave the "::[3]" in the name since that is the way you can indicate this
test specifically from the pytest command line. [This is perhaps because
the modifyitems hook is only called after test selection.]

*pytest_ignore_collect* skips kernelcl.py if pyopencl cannot be imported.
"""
from __future__ import print_function

import os.path
import inspect

import pytest
from _pytest.unittest import TestCaseFunction

# CRUFT: pytest<5.4 does not have from_parent class method.
function_test = getattr(pytest.Function, 'from_parent', pytest.Function)

def pytest_pycollect_makeitem(collector, name, obj):
    """
    Convert test generator into list of function tests so that pytest doesn't
    complain about deprecated yield tests.

    Note that unlike nose, the tests are generated and saved instead of run
    immediately.  This means that any dynamic context, such as a for-loop
    variable, must be captured by wrapping the yield result in a function call.

    For example::

        for value in 1, 2, 3:
            for test in test_cases:
                yield test, value

    will need to be changed to::

        def build_test(test, value):
            return test, value
        for value in 1, 2, 3:
            for test in test_cases:
                yield build_test(test, value)

    This allows the context (test and value) to be captured by lexical closure
    in build_test. See https://stackoverflow.com/a/233835/6195051.
    """
    if collector.istestfunction(obj, name) and is_generator(obj):
        tests = []
        for number, yielded in enumerate(obj()):
            index, call = split_yielded_test(yielded, number)
            description = getattr(call, 'description', name+index)
            test = function_test(
                parent=collector, name=description, callobj=call)
            tests.append(test)
        return tests

def is_generator(func):
    """
    Returns True if function has yield.
    """
    # Cribbed from _pytest.compat is_generator and iscoroutinefunction; these
    # may not be available as part of pytest 4.
    coroutine = (getattr(func, '_is_coroutine', False) or
                 getattr(inspect, 'iscoroutinefunction', lambda f: False)(func))
    generator = inspect.isgeneratorfunction(func)
    return generator and not coroutine

def split_yielded_test(obj, number):
    if not isinstance(obj, (tuple, list)):
        obj = (obj,)
    assert callable(obj[0])
    index = "[%d]"%number
    call = obj[0]
    return index, call

USE_DOCSTRING_AS_DESCRIPTION = True
def pytest_collection_modifyitems(session, config, items):
    """
    Add description to the test node id if item is a function and function
    has a description attribute or __doc__ attribute.
    """
    for item in items:
        if isinstance(item, pytest.Function):
            if isinstance(item, TestCaseFunction):
                # TestCase uses item.name to find the method so skip
                continue
            function = item.obj

            # If the test case provides a "description" attribute then use it
            # as an extended description.  If there is no description attribute,
            # then perhaps use the test docstring.
            if USE_DOCSTRING_AS_DESCRIPTION:
                description = getattr(function, 'description', function.__doc__)
            else:
                description = getattr(function, 'description', "")

            # If description is not supplied but yield args are, then use the
            # yield args for the description
            if not description and getattr(item, '_args', ()):
                description = (str(item._args) if len(item._args) > 1
                               else str(item._args[0]))

            # Set the description as part of the node identifier.
            if description:
                # Strip spaces from start and end and strip dots from end
                # pytest converts '.' to '::' on output for some reason.
                description = description.strip().rstrip('.')
                # Join multi-line descriptions into a single line
                if '\n' in description:
                    description = " ".join(line.strip()
                                           for line in description.split('\n'))

                # Note: leave the current name mostly as-is since the prefix
                # is needed to specify the nth test from a list of tests.
                item.name += "::" + description



# rectangular_prism model
# Note: model title and parameter table are inserted automatically
r"""
Definition
----------
This model provides the form factor, $P(q)$, for a hollow rectangular prism
with infinitely thin walls. It computes only the 1D scattering, not the 2D.
The 1D scattering intensity for this model is calculated according to the
equations given by Nayuk and Huber\ [#Nayuk2012]_.

Assuming a hollow parallelepiped with infinitely thin walls, edge lengths
$A \le B \le C$ and presenting an orientation with respect to the
scattering vector given by $\theta$ and $\phi$, where $\theta$ is the angle
between the $z$ axis and the longest axis of the parallelepiped $C$, and
$\phi$ is the angle between the scattering vector (lying in the $xy$ plane)
and the $y$ axis, the form factor is given by

.. math::

    P(q) = \frac{1}{V^2} \frac{2}{\pi} \int_0^{\frac{\pi}{2}}
           \int_0^{\frac{\pi}{2}} [A_L(q)+A_T(q)]^2 \sin\theta\,d\theta\,d\phi

where

.. math::

    V &= 2AB + 2AC + 2BC \\
    A_L(q) &=  8 \times \frac{
            \sin \left( \tfrac{1}{2} q A \sin\phi \sin\theta \right)
            \sin \left( \tfrac{1}{2} q B \cos\phi \sin\theta \right)
            \cos \left( \tfrac{1}{2} q C \cos\theta \right)
        }{q^2 \, \sin^2\theta \, \sin\phi \cos\phi} \\
    A_T(q) &=  A_F(q) \times
      \frac{2\,\sin \left( \tfrac{1}{2} q C \cos\theta \right)}{q\,\cos\theta}

and

.. math::

  A_F(q) =  4 \frac{ \cos \left( \tfrac{1}{2} q A \sin\phi \sin\theta \right)
                       \sin \left( \tfrac{1}{2} q B \cos\phi \sin\theta \right) }
                     {q \, \cos\phi \, \sin\theta} +
              4 \frac{ \sin \left( \tfrac{1}{2} q A \sin\phi \sin\theta \right)
                       \cos \left( \tfrac{1}{2} q B \cos\phi \sin\theta \right) }
                     {q \, \sin\phi \, \sin\theta}

The 1D scattering intensity is then calculated as

.. math::

  I(q) = \text{scale} \times V \times (\rho_\text{p} - \rho_\text{solvent})^2 \times P(q)

where $V$ is the surface area of the rectangular prism, $\rho_\text{p}$
is the scattering length density of the parallelepiped, $\rho_\text{solvent}$
is the scattering length density of the solvent, and (if the data are in
absolute units) *scale* is related to the total surface area.

**The 2D scattering intensity is not computed by this model.**


Validation
----------

Validation of the code was conducted  by qualitatively comparing the output
of the 1D model to the curves shown in (Nayuk, 2012\ [#Nayuk2012]_).


References
----------

See also Onsager [#Onsager1949]_.

.. [#Nayuk2012] R Nayuk and K Huber, *Z. Phys. Chem.*, 226 (2012) 837-854

.. [#Onsager1949] L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** Miguel Gonzales **Date:** February 26, 2016
* **Last Modified by:** Paul Kienzle **Date:** October 15, 2016
* **Last Reviewed by:** Paul Butler **Date:** September 07, 2018
"""

import numpy as np
from numpy import inf

name = "hollow_rectangular_prism_thin_walls"
title = "Hollow rectangular parallelepiped with thin walls."
description = """
    I(q)= scale*V*(sld - sld_solvent)^2*P(q)+background
        with P(q) being the form factor corresponding to a hollow rectangular
        parallelepiped with infinitely thin walls.
"""
category = "shape:parallelepiped"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld", "1e-6/Ang^2", 6.3, [-inf, inf], "sld",
               "Parallelepiped scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["length_a", "Ang", 35, [0, inf], "volume",
               "Shorter side of the parallelepiped"],
              ["b2a_ratio", "Ang", 1, [0, inf], "volume",
               "Ratio sides b/a"],
              ["c2a_ratio", "Ang", 1, [0, inf], "volume",
               "Ratio sides c/a"],
             ]

source = ["lib/gauss76.c", "hollow_rectangular_prism_thin_walls.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume", "equivalent outer volume sphere",
    "half length_a", "half length_b", "half length_c",
    "equivalent outer circular cross-section",
    "half ab diagonal", "half diagonal",
    ]


def random():
    """Return a random parameter set for the model."""
    a, b, c = 10**np.random.uniform(1, 4.7, size=3)
    pars = dict(
        length_a=a,
        b2a_ratio=b/a,
        c2a_ratio=c/a,
    )
    return pars

tests = [[{}, 0.2, 0.837719188592],
         [{}, [0.2], [0.837719188592]],
        ]
